/*	Maze LR2 Level Creator
	by Anthony M Kozar Jr
	Derived from SquareLR.cp
	Begun 6/17/03
	
	5/30/04: First working version completed.
	2/18/05: Added #ifdefs for Mac-specific code.
 */

#include	<iostream>
#include	<vector>
#include	<string>
#include	<fstream>
#include	<algorithm>

#if defined(macintosh) && defined(__MWERKS__)	/* CodeWarrior Mac compiler */
#include	"unix.h"
#include	"SIOUX.h"
#endif

#include	"lr2.h"
#include	"mazegen.h"

using namespace	std;
using	namespace	lr2;


/*	Function Prototypes    */

void	InitConsole();
void	AddExit( maze_type<simple_block>* maze );
void	WriteLevel(ofstream &fileout, maze_type<simple_block>* maze);
void	Header(ofstream &fileout, world_t world, int16 backgndID = (int16)0x2af8);
void	Monks(ofstream &fileout);
void	Items(ofstream &fileout);
void	MazeItems(ofstream &fileout, maze_type<simple_block>* maze);
void	MazeLayout(ofstream &fileout, maze_type<simple_block>* maze);
void	Gold(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor);
void	EntryPoint(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor);
void	WriteCoord(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor);
void	YHeader(ofstream &fileout, int16 Xnum, int16 Ycor);
void	ZHeader(ofstream &fileout, int16 Ynum, int16 Zcor);
void	Block(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor/*, block_t theBlock*/);
void	ExitBlock(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor/*, block_t theBlock*/);
void	SqrItems(ofstream &fileout, int sqr_size);
void	SquareLayout(ofstream &fileout, int sqr_size);
//void	Divider(ofstream &fileout, int length);
//void	Terminator(ofstream &fileout, int length);

/*	Constants   */

const	int	PLATFORM_Z		= 100;
const	int	MAZE_Z		= 99;
const	int	ENTRY_Z		= 95;

/*	Main Program    */

int	main()
{
	int	size;
	string filename;
	ofstream fout;
	maze_type<simple_block>* maze = NULL;
	
	InitConsole();
	
	maze = MakeOneMaze();
	AddExit(maze);
	maze->print_maze(cout);
	
	cout	<< "Enter a filename to save as: " << endl;
	cin	>> filename;

	fout.open(filename.c_str(), ios_base::out | ios_base::binary); 	// open file for binary output
	if (! fout)	{
		cout	<< "Could not open the output file.  Exiting." << endl;
		return -1;
	}
	
	WriteLevel(fout, maze);
	fout.close();

	return 0;
}

void	InitConsole()
{
#if defined(macintosh) && defined(__MWERKS__)	/* CodeWarrior Mac compiler */
	// set output file type to LR2 level
	_fcreator = 'LR2E';
	_ftype = 'LEVL';
	
	// don't prompt to save console window
	SIOUXSettings.asktosaveonclose = false;
#endif
	
	return;
}

void	AddExit( maze_type<simple_block>* maze )
{
	coord	exit;
	int	columns = maze->xsize;
	int	rows = maze->ysize;
	
	// choose where to put the exit
	bool	edge = rand() % 4;		// choose an edge of the maze
	switch (edge)	{
		case	0:				// left column
			exit.first = 1;
			exit.second = rand()%rows + 1;
			break;
		case	1:				// bottom row
			exit.first = rand()%columns + 1;
			exit.second = rows;
			break;
		case	2:				// right column
			exit.first = columns;
			exit.second = rand()%rows + 1;
			break;
		case	3:				// or top row
			exit.first = rand()%columns + 1;
			exit.second = 1;
			break;
	}
	
	maze->block_type(exit, kMExit);
	return;	
}

void	WriteLevel(ofstream &fileout, maze_type<simple_block>* maze)
{
	cout  << "writing ... ";
	Header( fileout, kBuilderWorld );
	Monks(fileout);
	MazeItems(fileout,maze);
	//Divider(fileout,180);
	MazeLayout(fileout,maze);
	//Terminator(fileout,222);
	cout  << "done." << endl;
	
	return;
}

void	Header(ofstream &fileout, world_t world, int16 backgndID)
{
	//using	namespace	lr2;
	
	cout  << "header, ";
	//fileout << LEVEL_FILE_ID;			// "magic" file ID
	fileout << int16(0xD87E) << int16(0x1B17);
	
	// I don't really know the purpose of the first part of the header
	if	( world == kWackyWorld )		// Wacky world header is different
		for	(int i = 0; i < H1_SIZE; i++)	fileout << WACKY_HEADER[i];
	else	for	(int i = 0; i < H1_SIZE; i++)	fileout << HEADER_PART_1[i];
	
	// values for the world and background of the level
	fileout << world;
	fileout << (int16)0x0000;
	fileout << backgndID;
	
	// the end of the header is all zeros unless scroll-locking is enabled
	for	(int i = 0; i < H2_SIZE; i++)	fileout << HEADER_PART_2[i];
	return;
}

void	Monks(ofstream &fileout)
{
	//using	namespace	lr2;
	
	cout  << "no monks, ";
//	for	(int i = 0; i < 4; i++)		fileout << monk_header[i];
	fileout << int16(0x0000);
	fileout << int16(0x0005);
	return;
}

void	Items(ofstream &fileout)
{
	cout  << "items, ";
	/*restore item_data
	for i = 1 to 32
	read value
	fileout << (char)value;
	next i
	*/	
	return;
}

void	MazeItems(ofstream &fileout, maze_type<simple_block>* maze)
{
	cout  << "items, ";

	// item block header
	fileout << int16(1);							// one item (the LR entry)
	fileout << int16(3);							// identifies this as beginning of items

	// find entry point
	int	xsize = maze->xsize,
		ysize = maze->ysize;
		
	for	(int y = kMinimumY; y <= (ysize+kMinimumY-1); y++) {
		for	(int x = (xsize+kMinimumX-1); x >= kMinimumX; x--)	{
			if	(maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMStart)	{
				EntryPoint(fileout, x, y, ENTRY_Z);			// add it
				goto done;
			}
		}
	}
	
	done:
	return;
}

void	MazeLayout(ofstream &fileout, maze_type<simple_block>* maze)
{
	cout  << "layout, ";

	int	currentZ = kMinimumZ;
	for	( ; currentZ < MAZE_Z; currentZ++ )
		fileout << int16(0xFFFF);						// "skip this Z plane"
		
	// add maze blocks
	int	xsize = maze->xsize,
		ysize = maze->ysize,	
		xnum;
		
	ZHeader(fileout, ysize, currentZ);
	for	(int y = kMinimumY; y <= (ysize+kMinimumY-1); y++) {
		xnum = 0;
		for	(int x = (xsize+kMinimumX-1); x >= kMinimumX; x--) {	// count the blocks for this Y
			if	(maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMWall)	++xnum;
			if	(maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMExit)	++xnum;
		}
		YHeader(fileout, xnum, y);
		for	(int x = (xsize+kMinimumX-1); x >= kMinimumX; x--) {	// add the blocks for this Y
			if	(maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMWall)
				Block(fileout,x,y,currentZ);				// use default block type
			else	if	(maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMExit)
				ExitBlock(fileout,x,y,currentZ);			// exit block
		}
	}
	++currentZ;
	
	// make platfrom - a rectangle of blocks
	ZHeader(fileout, ysize, currentZ);
	for	(int y = kMinimumY; y <= (ysize+kMinimumY-1); y++) {
		YHeader(fileout, xsize, y);
		for	(int x = (xsize+kMinimumX-1); x >= kMinimumX; x--)	{
			Block(fileout,x,y,currentZ);					// use default block type
		}
	}
	++currentZ;
	
	for	( ; currentZ <= kMaximumZ; currentZ++ )
		fileout << int16(0xFFFF);						// "skip this Z plane"
				
	return;
}


inline void	EntryPoint(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor)
{
	fileout << int16(0x00A2);					// entry point sequence
	fileout << int16(0x0003);
	fileout << int16(0xFFFF);
	WriteCoord(fileout, Xcor, Ycor, Zcor);
	return;
}

inline void	Gold(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor)
{
	fileout << int16(0x00AD);					// gold piece sequence
	fileout << int16(0x0001);
	fileout << int16(0xFFFF);
	WriteCoord(fileout, Xcor, Ycor, Zcor);
	return;
}

inline void	ZHeader(ofstream &fileout, int16 Ynum, int16 Zcor)
{
	//cout  << "   Z";

	fileout << Ynum;
	fileout << int16(0x0003);
	fileout << Zcor;
	return;
}

inline void	YHeader(ofstream &fileout, int16 Xnum, int16 Ycor)
{
	//cout  << " Y";
	
	fileout << Xnum;
	fileout << int16(0x000A);
	fileout << Ycor;
	return;
}

void	Block(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor/*, block_t theBlock*/)
{
	//cout  << "+";

	for	(int i = 0; i < B1_SIZE; i++)	fileout << DIGGABLE_BLOCK1[i];
	WriteCoord(fileout,Xcor,Ycor,Zcor);
	for	(int i = 0; i < B2_SIZE; i++)	fileout << DIGGABLE_BLOCK2[i];
	fileout << int16(0x1389);					// basic "builder world" block

	return;
}

void	ExitBlock(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor/*, block_t theBlock*/)
{
	//cout  << "E";

	for	(int i = 0; i < E1_SIZE; i++)	fileout << EXIT_BLOCK1[i];
	WriteCoord(fileout,Xcor,Ycor,Zcor);
	for	(int i = 0; i < E2_SIZE; i++)	fileout << EXIT_BLOCK2[i];
	
	return;
}

inline void	WriteCoord(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor)
{
	fileout << Xcor;
	fileout << Ycor;
	fileout << Zcor;
	return;
}
		
				

void	SqrItems(ofstream &fileout, int sqr_size)
{
	cout  << "items, ";
	
	// item block header
	fileout << int16(5);						// five items total
	fileout << int16(3);						// identifies this as beginning of items
	
	Gold(fileout, sqr_size+99, 100, 99);			// four gold pieces at corners
	Gold(fileout, 100, 100, 99);
	Gold(fileout, sqr_size+99, sqr_size+99, 99);
	Gold(fileout, 100, sqr_size+99, 99);
	EntryPoint(fileout, int(100 + sqr_size/2), 100, 97); 	// center on one edge of square
	
	return;
}

void	SquareLayout(ofstream &fileout, int sqr_size)
{
	int	sqr_center;
	cout  << "layout, ";
	
	// create an exit block in center of Z=99
	ZHeader(fileout, 1, 99);
	sqr_center = 100 + sqr_size/2;
	YHeader(fileout, 1, sqr_center);
	ExitBlock(fileout, sqr_center, sqr_center, int16(99));	// center it
	
	// make a square of blocks sqr_size on each side
	ZHeader(fileout,sqr_size, 100);
	for	(int y = 100; y <= (100 + sqr_size - 1); y++) {
		YHeader(fileout,sqr_size, y);
		for	(int x = (100 + sqr_size - 1); x >= 100; x--) {
			Block(fileout,x,y,100);					// use default block type
		}
	}
	
	return;
}

/*void	Divider(ofstream &fileout, int length)
{
	cout  << "divider, ";
	
	for	( int i = 0; i < length; i++ )
		fileout << (int8)255;

	return;
}

void	Terminator(ofstream &fileout, int length)
{
	cout  << "terminator. ";
	
	for	( int i = 0; i < length; i++ )
		fileout << (int8)255;

	return;
}

*/