/*	Maze LR2 Level Creator
	by Anthony M Kozar Jr
	Derived from SquareLR.cp
	Begun 6/17/03
	
	5/30/04: First working version completed.
	2/18/05: Added #ifdefs for Mac-specific code.
	5/12/05: Version 1.0
	         Added batch generation of levels, menu system, license info, log file.
	         Added random seeding and fixed several problems with exit placement.
	5/13/05: Version 1.1
	         Reversed Entrance/Exit to make levels more challenging.
 */

#include	<iostream>
#include	<vector>
#include	<string>
#include	<fstream>
#include	<algorithm>
#include	<cstdlib>
#include	<ctime>
#include	<cmath>

#if defined(macintosh) && defined(__MWERKS__)	/* CodeWarrior Mac compiler */
#include	"unix.h"
#include	"SIOUX.h"
#endif

#include	"lr2.h"
#include	"mazegen.h"

using namespace	std;
using	namespace	lr2;


/*	Function Prototypes    */

void	InitConsole();
void	SetTextFileType();
void	SetLevelFileType();
void	PrintLicense();
int	MakeOne();
int	MakeBatchOfMazes();
coord	FindExit(maze_type<simple_block>* maze);
void	AddEntrance( maze_type<simple_block>* maze );
void	WriteLevel(ofstream &fileout, maze_type<simple_block>* maze);
void	Header(ofstream &fileout, world_t world, int16 backgndID = (int16)0x2af8);
void	Monks(ofstream &fileout);
void	Items(ofstream &fileout);
void	MazeItems(ofstream &fileout, maze_type<simple_block>* maze);
void	MazeLayout(ofstream &fileout, maze_type<simple_block>* maze);
void	Gold(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor);
void	EntryPoint(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor);
void	WriteCoord(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor);
void	YHeader(ofstream &fileout, int16 Xnum, int16 Ycor);
void	ZHeader(ofstream &fileout, int16 Ynum, int16 Zcor);
void	Block(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor/*, block_t theBlock*/);
void	ExitBlock(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor/*, block_t theBlock*/);
void	SqrItems(ofstream &fileout, int sqr_size);
void	SquareLayout(ofstream &fileout, int sqr_size);
//void	Divider(ofstream &fileout, int length);
//void	Terminator(ofstream &fileout, int length);

/*	Constants   */

const	int	PLATFORM_Z		= 100;
const	int	MAZE_Z		= 99;
const	int	ENTRY_Z		= 95;

const	char	VERSION[]		= "1.1";

const	char	LICENSE_TEXT[]	=
	"Maze Generator\n"
	"Copyright (c) 2005 Anthony M. Kozar Jr.\n"
	"\n"
	"Redistribution and use in source and binary forms, with or without\n"
	"modification, are permitted provided that the following conditions are met:\n"
	"\n"
	"1.  Redistributions of source code must retain the above copyright\n"
	"    notice, this list of conditions and the following disclaimer. \n"
	"2.  Redistributions in binary form must reproduce the above copyright\n"
	"    notice, this list of conditions and the following disclaimer in the\n"
	"    documentation and/or other materials provided with the distribution. \n"
	"3.  The name of the author may not be used to endorse or promote\n"
	"    products derived from this software without specific prior written\n"
	"    permission. \n"
	"\n"
	"THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED\n"
	"WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF\n"
	"MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO\n"
	"EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,\n"
	"SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,\n"
	"PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;\n"
	"OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,\n"
	"WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR\n"
	"OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF\n"
	"ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n"
	"\n"
	"Maze Generator is not made by or supported by GT Interactive Software\n"
	"Corp., The WizardWorks Group, Inc., or any of such parties affiliates and\n"
	"subsidiaries.\n"
	"\n";

/*	Main Program    */

int	main()
{
	int	option;
	
	InitConsole();
	srand(time(NULL));
	
	cout	<< "Maze Generator " << VERSION << endl;
	cout	<< "by Anthony M. Kozar Jr." << endl;
	
	do	{
		cout	<< endl;
		cout	<< "Choose an option:" << endl
			<< "  (1)  Make one maze level" << endl
			<< "  (2)  Make a set of maze levels" << endl
			<< "  (3)  Print license information" << endl 
			<< "  (0)  Quit" << endl << "? ";
		cin	>> option;
		cout	<< endl;

		switch (option)	{
			case 0: break;
			case 1:
				if	(MakeOne())			option = 0;	// quit if make fails
				break;
			case 2:
				if	(MakeBatchOfMazes())	option = 0;	// quit if make fails
				break;
			case 3:
				PrintLicense();
				break;
			default:
				cout	<< "Please select one of the available choices.\n";
				break;
		}
	}
	while	(option && cin.good());
	
	cout	<< "Goodbye." << endl;
	return 0;
}

void	InitConsole()
{
#if defined(macintosh) && defined(__MWERKS__)	/* CodeWarrior Mac compiler */
	// don't prompt to save console window
	SIOUXSettings.asktosaveonclose = false;
	SIOUXSettings.rows = 48;
#endif
	
	SetLevelFileType();
	return;
}

void	SetTextFileType()
{
#if defined(macintosh) && defined(__MWERKS__)	/* CodeWarrior Mac compiler */
	// set output file type to LR2 level
	_fcreator = 'R*ch';
	_ftype = 'TEXT';
#endif
	
	return;
}

void	SetLevelFileType()
{
#if defined(macintosh) && defined(__MWERKS__)	/* CodeWarrior Mac compiler */
	// set output file type to LR2 level
	_fcreator = 'LR2E';
	_ftype = 'LEVL';
#endif
	
	return;
}


void	PrintLicense()
{
	cout	<< LICENSE_TEXT;
	return;
}

int	MakeOne()
{
	int	size;
	string filename;
	ofstream fout;
	maze_type<simple_block>* maze = NULL;
	
	cout	<< "Column and Row sizes must be odd integers within the range 3 to 200.\n\n";
	maze = MakeOneMaze();
	AddEntrance(maze);
	maze->print_maze(cout);
	
	cout	<< "Enter a filename to save as: " << endl;
	cin	>> filename;

	fout.open(filename.c_str(), ios_base::out | ios_base::binary); 	// open file for binary output
	if (! fout)	{
		cout	<< "Could not open the output file.  Exiting." << endl;
		return -1;
	}
	
	WriteLevel(fout, maze);
	fout.close();
	return 0;
}

int	MakeBatchOfMazes()
{
	int	numlevels, minrow, mincol, maxrow, maxcol, rows, columns, randomly, seed, digits;
	string fileprefix;
	char	filename[255];
	ofstream logfile;
	maze_type<simple_block>* maze = NULL;

	cout	<< "Column and Row sizes must be odd integers within the range 3 to 200.\n\n";
	cout	<< "How many maze levels? ";
	cin	>> numlevels;
	cout	<< "Minimum number of rows? ";
	cin	>> minrow;
	cout	<< "Maximum number of rows? ";
	cin	>> maxrow;
	cout	<< "Minimum number of columns? ";
	cin	>> mincol;
	cout	<< "Maximum number of columns? ";
	cin	>> maxcol;
	
	// check for validity of entered values
	if	(minrow < 0 or mincol < 0)	{
		cout	<< "Row and column sizes must be greater than zero!\n";
		return 0;
	}
	if	(maxrow < minrow or maxcol < mincol) {
		cout	<< "Maximum values cannot be smaller than minimum values!\n";
		return 0;
	}
	
	cout	<< "(1)  Randomly" << endl
		<< "(2)  Increase from minimum to maximum" << endl
		<< "How should the maze size be determined? ";
	cin	>> randomly;
	randomly = ! (randomly - 1);
	
	cout	<< "Enter a filename prefix: ";
	cin	>> fileprefix;
	
	cout	<< "Enter a seed value for the random number generator (0 for clock): ";
	cin	>> seed;
	
	if	(!seed)	seed = time(NULL);
	srand(seed);
	
	// open the log file and write parameters
	SetTextFileType();
	sprintf(filename, "%s%s\0", fileprefix.c_str(), ".log");
	logfile.open(filename);
	if (!logfile)	{
		cout	<< "Could not open log file" << filename << ".  Exiting." << endl;
		return -1;
	}
	logfile << filename << endl 
		  << "Levels created by Maze Generator" << endl
		  << "Version " << VERSION << endl << endl
		  << "Generating " << numlevels << " maze levels." << endl
		  << "Minimum rows:    " << minrow << endl
		  << "Maximum rows:    " << maxrow << endl
		  << "Minimum columns: " << mincol << endl
		  << "Maximum columns: " << maxcol << endl
		  << "Random seed:     " << seed   << endl << endl;
	
	SetLevelFileType();
	digits = (int)log10((double)numlevels) + 1;
	for	(int i = 1; i <= numlevels; i++)	{
		// calculate size
		if	(randomly)	{
			if	(minrow == maxrow) rows = minrow;
			else	rows = (rand() % (maxrow-minrow)) + minrow;
			if	(mincol == maxcol) columns = mincol;
			else	columns = (rand() % (maxcol-mincol)) + mincol;
		}
		else	{	// linearly interpolate between min and max
			rows = (((float)i/(float)numlevels) * (maxrow-minrow)) + minrow;
			columns = (((float)i/(float)numlevels) * (maxcol-mincol)) + mincol;
		}
		// make sure rows and columns are odd
		if	(!(rows%2))		++rows;
		if	(!(columns%2))	++columns;
		
		// create the maze
		maze = GenerateMaze(rows, columns);
		AddEntrance(maze);
		
		//cout	<< endl;
		//maze->print_maze(cout);					// print the maze to standard out
		//cout	<< endl;
		
		// determine filename, open, and write to file
		//string filename;
		ofstream fout;
		
		//filename = fileprefix + string(columns) + "x" + string(rows);
		sprintf(filename, "%s%.*d-%dx%d%s\0", fileprefix.c_str(), digits, i, columns, rows, ".lvl");
		fout.open(filename, ios_base::out | ios_base::binary); 	// open file for binary output
		if (! fout)	{
			cout	<< "Could not open output file" << filename << ".  Exiting." << endl;
			return -1;
		}
		
		cout	<< "Writing " << columns << "x" << rows << " maze to file " << filename << ".\n";
		logfile << "Writing " << columns << "x" << rows << " maze to file " << filename << ".\n\n";
		maze->print_maze(logfile);					// print the maze to log file
		logfile << endl;
		
		WriteLevel(fout, maze);
		fout.close();
		delete maze;
		maze = NULL;
	}

	logfile.close();
	return 0;
}	

coord	FindExit(maze_type<simple_block>* maze)
{
	coord	loc;
	int	xsize = maze->xsize,
		ysize = maze->ysize;
		
	for	(loc.second = 1; loc.second <= ysize; loc.second++) {
		for	(loc.first = 1; loc.first <= xsize; loc.first++)	{
			if	(maze->block_type(coord(loc.first, loc.second)) == kMExit)	{
				return loc;
			}
		}
	}
	// should not get to this point
	cerr	<< "No exit block in maze!" << endl;
	exit(-1);
	
	return loc;
}

void	AddEntrance( maze_type<simple_block>* maze )
{
	coord	start, exit;
	bool	again;
	int	columns = maze->xsize;
	int	rows = maze->ysize;
	
	exit = FindExit(maze);
	// choose where to put the entrance
	do	{
		again = false;
		int	edge = rand() % 2;		// choose an edge of the maze
		switch (edge)	{
			case	0:				// left column
				start.first = 1;
				start.second = 2 * (rand()%(rows/2) + 1);	// only even numbered spots
				break;
			case	1:				// bottom row
				start.first = 2 * (rand()%(columns/2) + 1);
				start.second = rows;
				break;
			/* only want the entrance on the above edges */
			/*case	2:				// right column
				start.first = columns;
				start.second = 2 * (rand()%(rows/2) + 1);
				break;
			case	3:				// or top row
				start.first = 2 * (rand()%(columns/2) + 1);
				start.second = 1;
				break;*/
		}
		if	(start.first == exit.first and start.second == exit.second)
			again = true;			// can't have start and exit in the same spot
	}
	while	(again);
	
	maze->block_type(start, kMStart);
	return;	
}

void	WriteLevel(ofstream &fileout, maze_type<simple_block>* maze)
{
	//cout  << "writing ... ";
	Header( fileout, kBuilderWorld );
	Monks(fileout);
	MazeItems(fileout,maze);
	//Divider(fileout,180);
	MazeLayout(fileout,maze);
	//Terminator(fileout,222);
	//cout  << "done." << endl;
	
	return;
}

void	Header(ofstream &fileout, world_t world, int16 backgndID)
{
	//using	namespace	lr2;
	
	//cout  << "header, ";
	//fileout << LEVEL_FILE_ID;			// "magic" file ID
	fileout << int16(0xD87E) << int16(0x1B17);
	
	// I don't really know the purpose of the first part of the header
	if	( world == kWackyWorld )		// Wacky world header is different
		for	(int i = 0; i < H1_SIZE; i++)	fileout << WACKY_HEADER[i];
	else	for	(int i = 0; i < H1_SIZE; i++)	fileout << HEADER_PART_1[i];
	
	// values for the world and background of the level
	fileout << world;
	fileout << (int16)0x0000;
	fileout << backgndID;
	
	// the end of the header is all zeros unless scroll-locking is enabled
	for	(int i = 0; i < H2_SIZE; i++)	fileout << HEADER_PART_2[i];
	return;
}

void	Monks(ofstream &fileout)
{
	//using	namespace	lr2;
	
	//cout  << "no monks, ";
//	for	(int i = 0; i < 4; i++)		fileout << monk_header[i];
	fileout << int16(0x0000);
	fileout << int16(0x0005);
	return;
}

void	Items(ofstream &fileout)
{
	cout  << "items, ";
	/*restore item_data
	for i = 1 to 32
	read value
	fileout << (char)value;
	next i
	*/	
	return;
}

void	MazeItems(ofstream &fileout, maze_type<simple_block>* maze)
{
	//cout  << "items, ";

	// item block header
	fileout << int16(1);							// one item (the LR entry)
	fileout << int16(3);							// identifies this as beginning of items

	// find entry point
	int	xsize = maze->xsize,
		ysize = maze->ysize;
		
	for	(int y = kMinimumY; y <= (ysize+kMinimumY-1); y++) {
		for	(int x = (xsize+kMinimumX-1); x >= kMinimumX; x--)	{
			if	(maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMStart)	{
				EntryPoint(fileout, x, y, ENTRY_Z);			// add it
				goto done;
			}
		}
	}
	
	done:
	return;
}

void	MazeLayout(ofstream &fileout, maze_type<simple_block>* maze)
{
	//cout  << "layout, ";

	int	currentZ = kMinimumZ;
	for	( ; currentZ < MAZE_Z; currentZ++ )
		fileout << int16(0xFFFF);						// "skip this Z plane"
		
	// add maze blocks
	int	xsize = maze->xsize,
		ysize = maze->ysize,	
		xnum;
		
	ZHeader(fileout, ysize, currentZ);
	for	(int y = kMinimumY; y <= (ysize+kMinimumY-1); y++) {
		xnum = 0;
		for	(int x = (xsize+kMinimumX-1); x >= kMinimumX; x--) {	// count the blocks for this Y
			if	(maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMWall)	++xnum;
			if	(maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMExit)	++xnum;
		}
		YHeader(fileout, xnum, y);
		for	(int x = (xsize+kMinimumX-1); x >= kMinimumX; x--) {	// add the blocks for this Y
			if	(maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMWall)
				Block(fileout,x,y,currentZ);				// use default block type
			else	if	(maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMExit)
				ExitBlock(fileout,x,y,currentZ);			// exit block
		}
	}
	++currentZ;
	
	// make platfrom - a rectangle of blocks
	ZHeader(fileout, ysize, currentZ);
	for	(int y = kMinimumY; y <= (ysize+kMinimumY-1); y++) {
		YHeader(fileout, xsize, y);
		for	(int x = (xsize+kMinimumX-1); x >= kMinimumX; x--)	{
			Block(fileout,x,y,currentZ);					// use default block type
		}
	}
	++currentZ;
	
	for	( ; currentZ <= kMaximumZ; currentZ++ )
		fileout << int16(0xFFFF);						// "skip this Z plane"
				
	return;
}


inline void	EntryPoint(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor)
{
	fileout << int16(0x00A2);					// entry point sequence
	fileout << int16(0x0003);
	fileout << int16(0xFFFF);
	WriteCoord(fileout, Xcor, Ycor, Zcor);
	return;
}

inline void	Gold(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor)
{
	fileout << int16(0x00AD);					// gold piece sequence
	fileout << int16(0x0001);
	fileout << int16(0xFFFF);
	WriteCoord(fileout, Xcor, Ycor, Zcor);
	return;
}

inline void	ZHeader(ofstream &fileout, int16 Ynum, int16 Zcor)
{
	//cout  << "   Z";

	fileout << Ynum;
	fileout << int16(0x0003);
	fileout << Zcor;
	return;
}

inline void	YHeader(ofstream &fileout, int16 Xnum, int16 Ycor)
{
	//cout  << " Y";
	
	fileout << Xnum;
	fileout << int16(0x000A);
	fileout << Ycor;
	return;
}

void	Block(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor/*, block_t theBlock*/)
{
	//cout  << "+";

	for	(int i = 0; i < B1_SIZE; i++)	fileout << DIGGABLE_BLOCK1[i];
	WriteCoord(fileout,Xcor,Ycor,Zcor);
	for	(int i = 0; i < B2_SIZE; i++)	fileout << DIGGABLE_BLOCK2[i];
	fileout << int16(0x1389);					// basic "builder world" block

	return;
}

void	ExitBlock(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor/*, block_t theBlock*/)
{
	//cout  << "E";

	for	(int i = 0; i < E1_SIZE; i++)	fileout << EXIT_BLOCK1[i];
	WriteCoord(fileout,Xcor,Ycor,Zcor);
	for	(int i = 0; i < E2_SIZE; i++)	fileout << EXIT_BLOCK2[i];
	
	return;
}

inline void	WriteCoord(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor)
{
	fileout << Xcor;
	fileout << Ycor;
	fileout << Zcor;
	return;
}
		
				

void	SqrItems(ofstream &fileout, int sqr_size)
{
	cout  << "items, ";
	
	// item block header
	fileout << int16(5);						// five items total
	fileout << int16(3);						// identifies this as beginning of items
	
	Gold(fileout, sqr_size+99, 100, 99);			// four gold pieces at corners
	Gold(fileout, 100, 100, 99);
	Gold(fileout, sqr_size+99, sqr_size+99, 99);
	Gold(fileout, 100, sqr_size+99, 99);
	EntryPoint(fileout, int(100 + sqr_size/2), 100, 97); 	// center on one edge of square
	
	return;
}

void	SquareLayout(ofstream &fileout, int sqr_size)
{
	int	sqr_center;
	cout  << "layout, ";
	
	// create an exit block in center of Z=99
	ZHeader(fileout, 1, 99);
	sqr_center = 100 + sqr_size/2;
	YHeader(fileout, 1, sqr_center);
	ExitBlock(fileout, sqr_center, sqr_center, int16(99));	// center it
	
	// make a square of blocks sqr_size on each side
	ZHeader(fileout,sqr_size, 100);
	for	(int y = 100; y <= (100 + sqr_size - 1); y++) {
		YHeader(fileout,sqr_size, y);
		for	(int x = (100 + sqr_size - 1); x >= 100; x--) {
			Block(fileout,x,y,100);					// use default block type
		}
	}
	
	return;
}

/*void	Divider(ofstream &fileout, int length)
{
	cout  << "divider, ";
	
	for	( int i = 0; i < length; i++ )
		fileout << (int8)255;

	return;
}

void	Terminator(ofstream &fileout, int length)
{
	cout  << "terminator. ";
	
	for	( int i = 0; i < length; i++ )
		fileout << (int8)255;

	return;
}

*/