/*	simple_block.h	*/

#ifndef	ANTHONY_SIMPLE_BLOCK_H
#define	ANTHONY_SIMPLE_BLOCK_H

#include	<utility>								// standard library file containing class pair

namespace	anthony_sblock	{

class	simple_block
{
	public:
		typedef std::pair<unsigned int, unsigned int> coord;	// define our coordinate type
		enum	block_t	{ clear, wall, start, exit, gold };
		enum	direction_t	{ north, east, south, west };		// define our direction type
		static const	direction_t	first_dir = north;
		static const	direction_t last_dir  = west;
		
		simple_block();
		bool	passable(direction_t direction) const;
		bool	crumbed() const	{ return _crumbed; }
		
		void	set(block_t value)	{ _contents = value; }
		block_t contents() const 	{ return _contents; }
		
		void	leave_crumb()	{ _crumbed = true; }
		void	remove_crumb()	{ _crumbed = false; }
		
	private:
	
		block_t	_contents;
		bool		_crumbed;
};

// constructor
inline simple_block::simple_block()
{
	_contents = clear;
	_crumbed = false;
}

// member functions

inline bool	simple_block::passable(direction_t direction) const
{
	direction = north;					// direction doesn't matter, this statement avoids compiler warning
	if	( _contents == clear )
		return true;
	else	return false;
}

// Overloaded Increment Operators for direction_t 
// since directions are an unordered set, these functions rotate thru the possible values without end
simple_block::direction_t& operator++( simple_block::direction_t& which_way );	// prefix ++ operator
simple_block::direction_t operator++( simple_block::direction_t& which_way, int );	// postfix ++ operator
			

}	// closure of namespace

#endif
