/*	stack2.h (version 1.1)
	
	A simple stack template class using a linked list implementation.  Member functions
	include the bare push(), pop(), and empty().  While it has the same interface as 
	the stack in stack.h, the size parameter to the constructor does not matter, but is 
	provided for compatibility.  The push() method should only fail if the application 
	heap runs out of space.  It now catches bad_alloc exceptions and returns false.
 */

#ifndef	ANTHONY_STACK2_H
#define	ANTHONY_STACK2_H

namespace	anthony_stack2	{

template	<class Type>
class	stack
{
	public:
	
		typedef unsigned int	size_t;
		static const size_t DEFAULT_SIZE = 50;
		
		stack( size_t size = DEFAULT_SIZE );
		~stack();
		
		bool	push(const Type& value);	// returns true only if successful
		bool	pop(Type& value); 		// value contains the popped item, returns true only if successful
		bool	empty();				// returns true if stack is empty
		
	private:
		template	<class Type>
		class	node						// basic node class to implement linked list
		{
			public:
				Type		value;
				node<Type>*	next;
		};
		
		node<Type>*	_top;					// pointer to the top of the stack

};

// constructor

template	<class Type>
stack<Type>::stack( size_t size )
{
	size = 0;							// ignore size in this implementation
	_top = NULL;						// an empty stack
}

// destructor

template	<class Type>
stack<Type>::~stack()
{
	node<Type>*	next;
	
	while	( _top != NULL )	{
		next = _top->next;
		delete _top;
		_top = next;
	}
}

// stack member functions

template	<class Type>
bool	stack<Type>::push(const Type& value)
{
try	{ 
	node<Type>*	new_top;

	new_top = new node<Type>;				// allocate a new top node
	new_top->value = value;					// store value
	new_top->next = _top;					// and pointer to old top
	_top = new_top;						// set _top to new node
	
	return true;
}
catch	(bad_alloc)	{ return false; }
}

template	<class Type>
bool	stack<Type>::pop(Type& value)
{
	node<Type>*	old_top;

	if	( empty() )
		return false;
	else	{
		old_top = _top;
		value = _top->value;				// get the top value of stack
		_top = _top->next;				// "pop" off the top node
		delete old_top;					// release memory for that node
		
		return true;
	}
}

template	<class Type>
inline bool	stack<Type>::empty()
{
	return ( _top == NULL );
}

		
}	// closure of namespace

#endif

