/*	Square LR2 Level Creator
	by Anthony M Kozar Jr
	written 11/11/99 in Chipmunk BASIC
	conversion to C++ started 12/17/00
	conversion finished 6/17/03
	
	2/18/05: Added #ifdefs for Mac-specific code
 */

#include	<iostream>
#include	<vector>
#include	<string>
#include	<fstream>
#include	<algorithm>

#if defined(macintosh) && defined(__MWERKS__)	/* CodeWarrior Mac compiler */
#include	"unix.h"
#include	"SIOUX.h"
#endif

#include	"lr2.h"

using namespace	std;
using	namespace	lr2;

void	InitConsole();

void	WriteLevel(ofstream &fileout, int sqr_size);
void	Header(ofstream &fileout);
void	Monks(ofstream &fileout);
void	Items(ofstream &fileout);
void	SqrItems(ofstream &fileout, int sqr_size);
void	Gold(ofstream &fileout, int8 Xcor, int8 Ycor, int8 Zcor);
void	EntryPoint(ofstream &fileout, int8 Xcor, int8 Ycor, int8 Zcor);
void	Divider(ofstream &fileout, int length = 180);
void	Terminator(ofstream &fileout, int length = 220);
void	WriteCoord(ofstream &fileout, int8 Xcor, int8 Ycor, int8 Zcor);
void	SquareLayout(ofstream &fileout, int sqr_size);
void	YHeader(ofstream &fileout, int8 Xnum, int8 Ycor);
void	ZHeader(ofstream &fileout, int8 Ynum, int8 Zcor);
void	Block(ofstream &fileout, int8 Xcor, int8 Ycor, int8 Zcor/*, block_t theBlock*/);
void	ExitBlock(ofstream &fileout, int8 Xcor, int8 Ycor, int8 Zcor/*, block_t theBlock*/);


int	main()
{
	int	size;
	string filename;
	ofstream fout;
	
	InitConsole();
	
	cout	<< "What size platform (3-100+)? " << endl;
	cin	>> size;
	cout	<< "Enter a filename to save as: " << endl;
	cin	>> filename;

	fout.open(filename.c_str(), ios_base::out | ios_base::binary); 	// open file for binary output
	if (! fout)	return -1;
	WriteLevel(fout, size);
	fout.close();

	return 0;
}

void	InitConsole()
{
#if defined(macintosh) && defined(__MWERKS__)	/* CodeWarrior Mac compiler */
	// set output file type to LR2 level
	_fcreator = 'LR2E';
	_ftype = 'LEVL';
	
	// don't prompt to save console window
	SIOUXSettings.asktosaveonclose = false;
#endif
	
	return;
}

void	WriteLevel(ofstream &fileout, int sqr_size)
{
	cout  << "writing ... ";
	Header(fileout);
	Monks(fileout);
	SqrItems(fileout,sqr_size);
	Divider(fileout,174);
	SquareLayout(fileout,sqr_size);
	Terminator(fileout);
	
	return;
}

void	Header(ofstream &fileout)
{
	//using	namespace	lr2;
	
	cout  << "header, ";
	for	(int i = 0; i < 34; i++)	fileout << header[i];
	return;
}

void	Monks(ofstream &fileout)
{
	//using	namespace	lr2;
	
	cout  << "monks, ";
	for	(int i = 0; i < 4; i++)		fileout << monk_header[i];
	return;
}

void	Items(ofstream &fileout)
{
	cout  << "items, ";
	/*restore item_data
	for i = 1 to 32
	read value
	fileout << (char)value;
	next i
	*/	
	return;
}

void	SqrItems(ofstream &fileout, int sqr_size)
{
	cout  << "items, ";
	
	fileout << (int8)0;						// item block header
	fileout << (int8)5;						// five items total
	fileout << (int8)0;
	fileout << (int8)3;						// identifies this as beginning of items
	
	Gold(fileout, sqr_size+99, 100, 99);			// four gold pieces at corners
	Gold(fileout, 100, 100, 99);
	Gold(fileout, sqr_size+99, sqr_size+99, 99);
	Gold(fileout, 100, sqr_size+99, 99);
	EntryPoint(fileout, int(100 + sqr_size/2), 100, 97); 	// center on one edge of square
	
	for	(int i = 0; i < 4; i++)	{				// item block terminator
		fileout << (int8)255;
	}
	return;
}

inline void	EntryPoint(ofstream &fileout, int8 Xcor, int8 Ycor, int8 Zcor)
{
	fileout << (int8)0;					// entry point sequence
	fileout << (int8)162;
	fileout << (int8)0;
	fileout << (int8)3;
	fileout << (int8)255;
	fileout << (int8)255;
	
	WriteCoord(fileout, Xcor, Ycor, Zcor);
	return;
}

inline void	Gold(ofstream &fileout, int8 Xcor, int8 Ycor, int8 Zcor)
{
	fileout << (int8)0;					// gold piece sequence
	fileout << (int8)173;
	fileout << (int8)0;
	fileout << (int8)1;
	fileout << (int8)255;
	fileout << (int8)255;
	
	WriteCoord(fileout, Xcor, Ycor, Zcor);
		
	return;
}

void	Divider(ofstream &fileout, int length)
{
	cout  << "divider, ";
	
	for	( int i = 0; i < length; i++ )
		fileout << (int8)255;

	return;
}

void	SquareLayout(ofstream &fileout, int sqr_size)
{
	int8	sqr_center;
	cout  << "layout, ";
	
	// create an exit block in center of Z=99
	ZHeader(fileout, 1, 99);
	sqr_center = int8(100 + sqr_size/2);
	YHeader(fileout, 1, sqr_center);
	ExitBlock(fileout, sqr_center, sqr_center, (int8)99);	// center it
	
	// make a square of blocks sqr_size on each side
	ZHeader(fileout,sqr_size, 100);
	for	(int y = 100; y <= (100 + sqr_size - 1); y++) {
		YHeader(fileout,sqr_size, y);
		for	(int x = (100 + sqr_size - 1); x >= 100; x--) {
			Block(fileout,x,y,100);			// use default block type
		}
	}
	
	return;
}

inline void	ZHeader(ofstream &fileout, int8 Ynum, int8 Zcor)
{
	//cout  << "   Z";

	fileout << (int8)0;
	fileout << Ynum;
	fileout << (int8)0;
	fileout << (int8)3;
	fileout << (int8)0;
	fileout << Zcor;
	
	return;
}

inline void	YHeader(ofstream &fileout, int8 Xnum, int8 Ycor)
{
	//cout  << " Y";
	
	fileout << (int8)0;
	fileout << Xnum;
	fileout << (int8)0;
	fileout << (int8)10;
	fileout << (int8)0;
	fileout << Ycor;
	
	return;
}

void	Block(ofstream &fileout, int8 Xcor, int8 Ycor, int8 Zcor/*, block_t theBlock*/)
{
/*	restore simple_block					// set next read point to data for a block
#	if block_type.first = 0 and block_type.second = 0
#	then let block_type.first = 19, block_type.second = 137	// default block type
	for i = 1 to 8
		read value
		fileout << (char)value;
	next i
	WriteCoord(fileout,Xcor,Ycor,Zcor)
	for i = 1 to 14
		read value
		fileout << (char)value;
	next i
	fileout << (char)19,  # fileout			// originally had block_type members here
	fileout << (char)137;			// but couldnt get them to work
*/	
	//cout  << "+";

	copy( block1, block1 + 8, ostream_iterator<int8>( fileout, "") );
	WriteCoord(fileout,Xcor,Ycor,Zcor);
	copy( block2, block2 + 14, ostream_iterator<int8>( fileout, "") );
	fileout << (int8)19;					// basic "builder world" block
	fileout << (int8)137;

	return;
}

void	ExitBlock(ofstream &fileout, int8 Xcor, int8 Ycor, int8 Zcor/*, block_t theBlock*/)
{
/*	restore exit_data						// set next read point to data for a block
	for i = 1 to 8
		read value
		fileout << (char)value;
	next i
	WriteCoord(fileout,Xcor,Ycor,Zcor)
	for i = 1 to 18
		read value
		fileout << (char)value;
	next i
*/
	//cout  << "E";

	copy( exit1, exit1 + 8, ostream_iterator<int8>( fileout, "") );
	WriteCoord(fileout,Xcor,Ycor,Zcor);
	copy( exit2, exit2 + 18, ostream_iterator<int8>( fileout, "") );
	
	return;
}

inline void	WriteCoord(ofstream &fileout, int8 Xcor, int8 Ycor, int8 Zcor)
{
	fileout << (int8)0;
	fileout << Xcor;
	fileout << (int8)0;
	fileout << Ycor;
	fileout << (int8)0;
	fileout << Zcor;
	
	return;
}
		
				

void	Terminator(ofstream &fileout, int length)
{
	cout  << "terminator. ";
	
	for	( int i = 0; i < length; i++ )
		fileout << (int8)255;

	return;
}

